#ifndef __NODE_H__
#define __NODE_H__

#include <assert.h>
#include "globals_catClus.h"
#include "globals.h"
#include "Set.h"


class Node
{
private:
  int _id;
  double _weight;

public:
  Node(): _id(-1), _weight(0.0){}
  Node(int id, double wt)
    {
      _id = id; 
      _weight = wt;
    }
  ~Node(){};
  
  void Init(int id, double wt)
    {
      _id = id;
      _weight = wt;
    }

  double Wt() { return _weight;}
  int Id() { return _id;}

  void SetWt(double val) { _weight = val;}
  void AddWt(double val) { _weight += val;}
  friend class Graph;
};


class Graph
{
private:
  int _numAttrs;
  int *_domSizes;
  int *_numBasins;
  Node *_nodes;
  CombinationOp _combineOp;

  // precompute and store the following;
  int *_startAttrVals; 
  int *_startNodeIndx;
  //int *_startBasinVals; 
  int _numNodes;
  int _totNumAttrVals;
  int _totNumBasins;

  // information about clusters on each attribute;
  int **_sortAttrValues;
  

  // just divides each component with its norm;
  void _NormalizeAttr(int attr, int basin);
  double _Norm(int attr, int basin);
  double _InnerProduct(int attr1, int basin1, int attr2, int basin2);

  // used for Gram-Schmidt orthonormalization
  // of the basin vectors of a particular attribute;
  void _NormalizeVec(int attr, int basin);
  void _ReduceFromVec(int attr, int basin1, int basin2, double factor);

  void _InitializeIDbasins(int attr);
  double _Combine(Set& tuple, int attr, int basin, double& overallVal)
    {
      int i;
      double val;
      
      switch(_combineOp)
	{
	case ADD:
	  val = 0.0;
	  if (0.0 == overallVal)
	    {
	      for (i = 0; i < tuple.size(); i++)
		{
		  assert(tuple.elements(i) < TotNumAttrVals());
		  if (attr != i)
		    val += GetWt(tuple.elements(i), i, basin);
		  //val += _nodes[tuple.elements(i)].Wt();
		}
	      overallVal = val + GetWt(tuple.elements(attr), attr, basin);
	    }
	  else
	    val = overallVal-GetWt(tuple.elements(attr), attr, basin);
	  break;
	case PRODUCT:
	  val = 1.0;
	  for (i = 0; i < tuple.size(); i++)
	    {
	      assert(tuple.elements(i) < TotNumAttrVals());
	      if (attr != i)
		val += GetWt(tuple.elements(i), i, basin);
	    }
	  break;
	case MAX:
	  assert(0);
	  break;
	default:
	  assert(0);
	}
      //cout << val << endl;
      return val;
    }

  //Revisit: change the mapping so that it can be
  // computed much faster;
  int _FindAttrVal(int cumVal)
    {
      int i;
      int sum = 0;
      for (i = 0; i < _numAttrs-1; i++)
	{
	  if (cumVal < sum+_domSizes[i+1])
	    return (cumVal - sum);
	  sum += _domSizes[i];
	}
      return (cumVal - _domSizes[_numAttrs-1]);
    }

public:
  Graph(int nA, int *domSz, int *basinSz, CombinationOp op=ADD);
  ~Graph();

  void ClearWts();
  int NumAttrs() { return _numAttrs;}
  int NumNodes() { return _numNodes;}
  int TotNumAttrVals() {return _totNumAttrVals;}
  int NumBasins(int attr) { return _numBasins[attr];}
  int DomainSz(int attr) { return _domSizes[attr];}
  int StartAttrVal(int attr) { return _startAttrVals[attr];}
  
  int StartNodeIndx2(int attr, int basin) 
    {
      return (_startNodeIndx[attr] + _domSizes[attr]*basin);
    }

  int StartNodeIndx(int attr)
    {
      return _startNodeIndx[attr];
    }

  double NodeWt(int i) {return  _nodes[i].Wt();}
    
  double GetWt(int attrVal, int attr, int basin)
    {
      //int aVal = _FindAttrVal(attrVal);
      int aVal;
      aVal = attrVal - _startAttrVals[attr];
      return _nodes[StartNodeIndx2(attr, basin)+aVal].Wt();
    }
  void SetWt(int attrVal, int attr, int basin, double val)
    {
      //int aVal = _FindAttrVal(attrVal);
      int aVal;
      aVal = (0 == attr) ? attrVal : attrVal - _startAttrVals[attr];
      _nodes[StartNodeIndx2(attr, basin)+aVal].SetWt(val);
    }

  void AddWtToNode(int attrVal, int attr, int basin, double addlWt)
    {
      //int correctVal = _FindAttrVal(attrVal);
      int aVal;
      aVal = (0 == attr) ? attrVal : attrVal - _startAttrVals[attr];
      _nodes[aVal + StartNodeIndx2(attr, basin)]._weight += addlWt;
    }

  void Initialize(InitialState init);

  void IterateSTIRR(char *inputFile, Graph *newGraph, PrefixTree *pt=NULL);
  void UpdateSTIRR(Set &tuple, Graph *graph2, int count=1);
  void OrthoNormalizeAllAttrs();
  void OrthoNormalizeAttr(int attr);
  
  void operator=(Graph& graph);
  void AssignAttr(Graph& graph, int attr);

  /**** Functions for post-processing an attribute ***/
  void Sort(int attr, int basinNum, int *sortAttrVals);
  // partitions the attribute values into a natural number of clusters
  int Partition(int attr, int basinNum, int *sortedAttrVals, int *breaks); 
  int OverlapClusters(int attr, int basinNum);

  void ClusterAttribute(int attr);

#ifdef DEBUG
  void TestOrthoNormality(int attr);
  void TestOrthoNormality();
#endif

  friend ofstream& operator<<(ofstream& of, Graph& graph);
  friend ostream& operator<<(ostream& of, Graph& graph);
};


#endif
