/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



// This file contains function main which is responsible for inputting the
// parameters for the execution of the algorithms, and for driving the
// algorithms.  The file also contains functions for counting 1- and
// 2-itemsets.  Counting of such small itemsets is done in an array,
// thus avoiding the overhead of Hash/Prefix tree manipulations.

#include <stdlib.h>
#include <fstream.h>

#include "Set.h"
#include "macros.h"
#include "srcFunctions.h"

#include "PrefixTree.h"

int count1and2Itemsets(
#ifdef SAMPLING
		       Set * sample[], int sampleSize,
#else
		       ifstream & xfile,
#endif
		       int * counts1, int * counts2,
		       int numItems, int minSupport) {
    int large2Itemsets = 0;

#ifdef SAMPLING
    for (int index = 0; index < sampleSize; index++) {
	Set transaction(*sample[index]);
#elif defined(ASCII)
    // Potentially very wasteful but we only need one buffer and set at a time.
    int strSize = MAX_TRANS_BYTES(numItems);
    char buf[strSize];

    xfile.getline(buf, strSize);

    while (xfile) {
	Set transaction(buf);
	xfile.getline(buf, strSize);
#else
    int xNumber = 0;
    xfile.read(&xNumber, sizeof(xNumber));

    while (xfile) {
	xfile.read(&xNumber, sizeof(xNumber));
	Set transaction(xfile);
	xfile.read(&xNumber, sizeof(xNumber));
#endif  // SAMPLING

	for (int i = 0; i < transaction.size(); i++) {
	    // Increment the 1-itemset counts.
	    counts1[transaction.elements(i)]++;

	    // Increment the two itemset counts.
	    for (int j = i + 1; j < transaction.size(); j++) {
		// Note that transaction[j] > transaction[i].  In this case we'll
		// only increment the count located in counts2[trans[i]][trans[j]],
		// and not in counts2[trans[j]][trans[i]] (conceptually).

		// LA: Verify the computation below.
		int k = transaction.elements(i) * numItems + transaction.elements(j);
		counts2[k]++;
		if (counts2[k] == minSupport) large2Itemsets++;
	    }
	}
    }

#if !defined(SAMPLING) && defined(ASCII)
    if (buf[0] != '\0')
	cerr << "no end of line before end of file in xfile" << endl
	     << "transaction " << buf << " ignored!" << endl;
#endif  // !SAMPLING && ASCII

    return large2Itemsets;
}

void usage(char * s) {
    cerr << "usage: " << s << "  -i <Xfile> -n <num_items> -m <min_sup>" << endl;
#ifdef SAMPLING
    for (int i = strlen(s); i; i--) cerr << ' ';

    cerr << "        -ss <sample_size> -l <low_min_sup>" << endl;
#endif

    for (int i = strlen(s); i; i--) cerr << ' ';

    cerr << "        [-s <sets_per_leaf> -h <hash_h> -a <hash_a> -b <hash_b>]" << endl
	 << "or     " << s << " -help" << endl << endl
	 << "Xfile \t\t- transaction file," << endl
	 << "num_items \t- positive integer denoting the number of items "
	 << "in the domain" << endl
	 << "min_sup \t- positive integer denoting the minimum support" << endl
#ifdef SAMPLING
	 << "sample_size \t- positive integer denoting the size of the sample" << endl
	 << "low_min_sup \t- positive integer denoting the lowered minimum support" << endl
#endif
	 << "sets_per_leaf \t- maximum number of itemsets to be stored in "
	 << " a leaf node of a hash tree" << endl
	 << "hash_h \t\t- parameters for the hash function" << endl
	 << "hash_a \t\t  ((hash_a) * (int) + (hash_b)) % (hash_h)" << endl
	 << "hash_b \t\t  hash_h and hash_a must > 0, hash_b must be >= 0" << endl;
}

void error(bool usage, char * s, char * s1, char * s2 = "") {
    cerr << s << ": " << s1 << ' ' << s2 << endl;
    if (usage) ::usage(s);
    exit(1);
}

PrefixTree* driverApriori(char *fileName, int numItems, int minSupport)
  {
    // Optional arguments.
    int setsPerLeaf = DEF_SETS_PER_LEAF;
    int h =DEF_HASH_H;
    int a =DEF_HASH_A;
    int b =DEF_HASH_B;

    // Required arguments.
    ifstream xfile;
    
#ifdef SAMPLING
    // Required arguments of the sampling algorithm.
    int sampleSize = 0;
    int lowMinSupport = 0;
#endif


#ifdef SAMPLING
    Set ** sample = new Set * [sampleSize];

    int read = readSample(xfile, numItems, sample, sampleSize);
    if (read != sampleSize) {
	char msg[80];
	sprintf(msg, "read only %d transactions into the sample from file", read);
	error(false, argv[0], msg, fileName);
    }
#endif

    // All argument input is done.  Now perform 1- and 2-itemset counting.

    int * counts1 = new int[numItems];             // Counts for the 1-itemsets.
    int * counts2 = new int[numItems * numItems];  // Counts for the 2-itemsets.

    // Clear the counts for the 1- and 2-itemsets.
    for (int i = 0; i < numItems; i++) {
	counts1[i] = 0;
	for (int j = 0; j < numItems; j++)
	    counts2[i * numItems + j] = 0;
    }

#ifdef SAMPLING
    count1and2Itemsets(sample, sampleSize, counts1, counts2,
		       numItems, lowMinSupport);
#else
    count1and2Itemsets(xfile, counts1, counts2, numItems, minSupport);
#endif

    Tree * T[numItems];

#ifdef HASH_TREE
    T[2] = new Tree(2, minSupport, setsPerLeaf, h, a, b);
    cout << "\nItemSet\t\t\t\t\tCount\n***********************************\n\n";

    for (int i = 0; i < numItems; i++) {
      // No need to look into counts2[i][?] if counts1[i] < minSupport
      if (counts1[i] >= minSupport) {
	cout << i << "\t\t\t\t\t" << counts1[i] << endl;
	for (int j = i + 1; j < numItems; j++) {
	  int count = counts2[i * numItems + j];
	  if (count >= minSupport) {
	    Item ij[] = {i, j};
	    Set s(2, ij);
	    T[2]->insert(s, count);
	  }
	}
      }
    }
#elif defined(SAMPLING)
    T[2] = new Tree(numItems, counts1, counts2, lowMinSupport, h, a, b);
#else
    T[2] = new Tree(numItems, counts1, counts2, minSupport, h, a, b);
#endif

    // Done with counts.
    delete[] counts1;
    delete[] counts2;

#ifdef APRIORI
    int maxSetLength = apriori(T, fileName, numItems);
#elif defined(SAMPLING)
    int maxSetLength = apriori(T, sample, sampleSize);

    // Don't need sample[] sets anymore.
    for (int i = 0; i < sampleSize; i++)
      delete sample[i];
    delete [] sample;
#endif  // APRIORI


#ifdef DIC_ALG
    // For now, dic will always return 0.
    dic(T, fileName, numItems);
#endif


#ifdef SAMPLING
    T[2]->setMinSupport(minSupport);
    countAndPrune(T[2], fileName, numItems);
#endif


    // Clean up.
    delete[] fileName;

    // Output and delete trees.
#ifdef HASH_TREE
    for (int i = 2; i <= maxSetLength; i++) {
      cout << *T[i];
      delete T[i];
    }
#else
    return T[2];    
#endif
}


