/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <assert.h>
#include <strstream.h>

#include "globals.h"
#include "Set.h"

// initialize maxNumItems to be size
// Can be used as the default (no argument) constructor.
Set::Set(int size = NUM_ITEM_STEP) {
  activeLow = 0;
  numItems = 0;

  // If numItems and maxNumItems are both 0, insertion
  // will not work.

  if (0 == size)
    items = NULL;
  else
    {
      maxNumItems = (size >= 0) ? size : NUM_ITEM_STEP;
      items = new Item[maxNumItems];
    }
}

// Constructor from an array
// Preconditions: numElements is at least 0,
// there numElements valid Items in a[],
// the items are sorted.
Set::Set(int numElements, const Item a[]) {
  activeLow = 0;
  numItems = numElements;

  //cout << " Inside Set::Set(num, int*) " << numElements << endl;

  // If numItems and maxNumItems are both 0, insertion
  // will not work.
  maxNumItems = (numElements > 0) ? numElements : NUM_ITEM_STEP;

  items = new Item[maxNumItems];

  // LA: Revisit: change to memcpy later.
  for (int index = 0; index < numItems; index++)
    items[index] = a[index];
}

// Preconditions: the first int in xfile contains the number of
// items in the transaction, and the items are sorted.
// Transactions in xfile begin with a transaction number but
// that is read by the caller as part of end of xfile test.
Set::Set(ifstream & xfile) {
    activeLow = 0;
    xfile.read(&numItems, sizeof(numItems));

	assert(numItems > 0);
    // If numItems and maxNumItems are both 0, insertion
    // will not work.
    maxNumItems = (numItems > 0) ? numItems : NUM_ITEM_STEP;
    items = new Item[maxNumItems];
	
    xfile.read(items, numItems * sizeof(Item));
}

// Preconditions: the items are sorted.
Set::Set(const char * s) {
    activeLow = 0;
    numItems = 0;
    maxNumItems = (int) ((strlen(s) + 1)/2);

    if (!maxNumItems) {
	// If numItems and maxNumItems are both 0, insertion
	// will not work.
	maxNumItems = NUM_ITEM_STEP;
	items = new Item[maxNumItems];
	return;
    }

    items = new Item[maxNumItems];

    istrstream istr(s);
    int i;

    istr >> i;
    while (istr) {
	items[numItems++] = i;
	istr >> i;
    }
}

// Copy constructor
Set::Set(const Set & s) {
  assert(!s.activeLow);
  activeLow = 0;
  numItems = s.numItems;

  // LA: Note our set will have maxNumItems equal
  // to the number of valid elements in s.  We may
  // avoid wasting space this way.
  // If numItems and maxNumItems are both 0, insertion
  // will not work.
  maxNumItems = (numItems > 0) ? numItems : s.maxNumItems;

  items = new Item[maxNumItems];

  // LA: Revisit: change to memcpy later.
  for (int index = 0; index < numItems; index++)
    items[index] = s.items[index];
}

//BEGIN VENKY
/**
Set::Set(int nitems, int *arr)
{
 activeLow = 0;
 numItems = nitems; 
 items = arr;
}
**/

int Set::NumBytes()
{
	return (sizeof(int) + numItems * sizeof(int));
}

int Set::newArr(int nitems, int *arr)
{
  activeLow = 0;
  numItems = nitems; 
  maxNumItems = nitems; 
  items = arr; 
  
  return (sizeof(int) + numItems * sizeof(Item));
}

void Set::DeleteArr()
{
  delete [] items;
  items = NULL;
}

void Set::Nullify()
{
  items = NULL;
}

void Set::WriteToFile(ofstream& ofile, int count)
{
	int END_MARKER = -1;

	ofile.write(&numItems, sizeof(int));
	ofile.write(items, numItems * sizeof(Item));
	ofile.write(&count, sizeof(int));
	ofile.write(&END_MARKER, sizeof(int));
}
//END VENKY

Set& Set::operator=(const Set & s) {
  assert(!(activeLow || s.activeLow));
  activeLow = 0;

  // Protect against s = s;
  if (items != s.items) {
    delete [] items;

    // LA: Note our set will have maxNumItems equal
    // to the number of valid elements in s.  We may
    // avoid wasting space this way.
    maxNumItems = numItems = s.numItems;

    items = new Item[maxNumItems];

    // LA: Revisit: change to memcpy later.
    for (int index = 0; index < numItems; index++)
      items[index] = s.items[index];
  }

  return *this;
}

Set::~Set() {
  assert(!activeLow);
  if (NULL != items)
  	delete [] items;
}

// insert item into *this, keep elements sorted
bool Set::insert(Item item) {
  assert(!activeLow);

  int insertPos = 0;

  for (; insertPos < numItems && items[insertPos] < item; insertPos++);

  if (insertPos < numItems && items[insertPos] == item)
    // No duplicate items should be inserted.
    return false;

  if (numItems == maxNumItems) {
    maxNumItems += NUM_ITEM_STEP;
    Item * newItems = new Item[maxNumItems];
    if (!newItems)
      return false;

    // Copy elements into the new array until we get to
    // insertPos.  Then copy the rest of the elements.
    // LA: Revisit: change to memcpy later.
    for (int index = 0; index <= numItems; index++)
      if (index < insertPos)
	newItems[index] = items[index];
      else if (index == insertPos)
	newItems[index] = item;
      else
	newItems[index] = items[index - 1];

    numItems++;
    delete [] items;
    items = newItems;
    return true;
  }

  // Shift elements down (from the end of the array) until
  // we get to insertPos.  Copy the item in that spot.
  // LA: Revisit: may be doable using memcpy, even with overlapping ptrs.
  for (int index = numItems; index > insertPos; index--)
    items[index] = items[index - 1];

  items[insertPos] = item;

  numItems++;
  return true;
}

// pull out item [start..end] and construct a new Set
// return pointer to this new set
// Preconditions: start..end represent a valid, non-empty
// subset of the set pointed to by this.
Set * Set::subset(int start, int end) const {
  assert(!activeLow);

  Set * result = new Set(end - start + 1);

  result->numItems = end - start + 1;
  // LA: Revisit: change to memcpy later.
  for (int index = start; index <= end; index++)
    result->items[index - start] = items[index];

  return result;
}

// construct a new Set that consists of elements of *this
// plus the new item;  return pointer to this new set
// The new item is appended at the end.  Set::insert()
// inserts an element in the current set, and preserves
// the sorted order.
Set * Set::addElement(Item item) const {
  assert(!activeLow);

  Set * result = new Set(numItems + 1);

  result->numItems = numItems + 1;
  // LA: Revisit: change to memcpy later.
  for (int index = 0; index < numItems; index++)
    result->items[index] = items[index];

  result->items[numItems] = item;

  return result;
}

// Yank out item k from this set and return this item
// No 'new' is involved
bool Set::yankOut(Item & i, int k) {
  assert(!activeLow);

  if (k < 0 || k >= numItems)
    return false;

  i = items[k];

  // LA: Revisit: may be doable using memcpy, even with overlapping ptrs.
  for (int index = k + 1; index < numItems; index++)
    items[index - 1] = items[index];

  numItems--;
  return true;
}

// should only be used in pair with and after yankOut
// As a result, e.g., maxNumItems shouldn't need to be increased
// insert this item, and item is currently not in the set.
// insert() is more general, and does not have the preconditions.
bool Set::putBack(Item item){
  assert(!activeLow);

  if (numItems == maxNumItems)
    return false;

  // Shift elements down (from the end of the array) until we
  // find the place to stick in item.  Copy the item in that spot.
  int index;
  // LA: Revisit: may be doable using memcpy, even with overlapping ptrs.
  for (index = numItems; index && items[index - 1] > item; index--)
    items[index] = items[index - 1];

  items[index] = item;

  numItems++;
  return true;
}

// return true if item is an element in this set
bool Set::member(Item item) const {
  assert(!activeLow);

  int index = 0;

  for (; index < numItems && items[index] < item; index++);

  if (index == numItems || items[index] != item)
    return false;
  else return true;
}

// Returns true iff *this equals s.
bool Set::operator==(const Set& s) const {
  assert(!(activeLow || s.activeLow));

  if (numItems != s.numItems)
    return false;

  // LA: Revisit: change to memcmp later.
  for (int index = 0; index < numItems; index++)
    if (items[index] != s.items[index])
      return false;

  return true;
}

//Venky: added the foll. function to make the subset join easy.
//Returns the element e which differentiates "this" from the second set s.
//That is, this union {e} = s;
//If not, return -1;


// Returns true iff *this is a subset of s.
bool Set::operator<=(const Set& s) const {
  assert(!(activeLow || s.activeLow));

  if (numItems > s.numItems)
    return false;

  // items in both sets are sorted, so avoid membership testing
  // for every element of *this.  Instead keep a search index
  // into s, so s gets traversed only once.
  int index2 = 0;
  for (int index = 0; index < numItems; index++) {
    while (index2 < s.numItems && items[index] != s.items[index2])
      index2++;
    
    if (index2 == s.numItems)
      return false;
  }
  return true;
}
	
// Used by LeafNode::sort().
// Lexicographical comparison of sets of the same size.
// The sizes of sets of different sizes are compared.
int compare(const Set & s1, const Set & s2) {
  assert(!(s1.activeLow || s2.activeLow));

  // LA: Should normally be called only for sets of the same
  // size (as part of sorting).  May want to remove the second
  // part of the test later for performance reasons, and make it
  // a precondition.  For now compare sizes of sets of different
  // sizes.  We test the first set for emptyness, since the loop
  // below assumes there is at least one valid element.
  if (!s1.numItems || s1.numItems != s2.numItems)
    return s1.numItems - s2.numItems;

  int index;
  for (index = 0; s1.items[index] == s2.items[index]; index++)
    if (index == s1.numItems - 1)
      return 0;

  return s1.items[index] - s2.items[index];    
}

ostream& operator<<(ostream& os, const Set& s) {
  //assert(!s.activeLow);

  os << "num = " << s.numItems << "  activeLow = " << s.activeLow << ": \t";

  //  os << '{';
  for (int index = 0; index < s.numItems; index++) {
    os << s.items[index];
    if (index < s.numItems - 1)
      os << ",";
  }

  return os;  // << '}' << endl;
}
