/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef ORDERED_SET
#define ORDERED_SET

#include "iostream.h"
#include "fstream.h"

#include "macros.h"
#include "globals.h"

class Set
{
  public:

	// initialize maxNumItems to be size
	// Can be used as the default (no argument) constructor.
	Set(int size = NUM_ITEM_STEP);

	// Constructor from an array
	Set(int numElements, const Item a[]);

	// Will be used to construct Sets out of transactions.
	Set(ifstream & xfile);

	// Will be used to construct Sets out of transactions
	// for testing and debugging purposes.
	Set(const char * s);

	// Copy constructor
	Set(const Set & s);

	// VENKY
	// the following sets the pointer to an array of items
	// without actually creating an array; Avoids the use 
	// of the new operator.
	// Set(int nitems, int *arr);

	int NumBytes();
	int newArr(int nitems, int *arr);
	void DeleteArr();
	void Nullify(); // set items to NULL;
	void WriteToFile(ofstream& ofile, int count);
	// END VENKY

	Set& operator=(const Set & s);

	~Set();

	// return numItems
	int size() const
	  { return numItems - activeLow; }

	// return the item with index index (starting from 0)
	Item elements(int index) const
	  { return items[activeLow + index]; }

	// insert item into *this, keep elements sorted
	bool insert(Item item);

	// pull out item [start..end] and construct a new Set
	// return pointer to this new set
	Set * subset(int start, int end) const;

	// Increments activeLow by its argument.  Used as a fast
	// subset with no new set construction.  Note that callers
	// should restore activeLow.  That can be done by passing
	// the negative of the original input to suffix.
	void suffix(int incStart)
	  { activeLow += incStart; }

	// construct a new Set that consists of elements of *this
	// plus the new item;  return pointer to this new set
	Set * addElement(Item item) const;

	// Yank out item k from this set and return this item
	// No 'new' is involved
	bool yankOut(Item & i, int k);

	// should only be used in pair with and after yankOut
	// As a result, e.g., maxNumItems shouldn't need to be increased
	// insert this item
	bool putBack(Item item);
	
	// return true if item is an element in this set
	bool member(Item item) const;

	// Returns true iff *this equals s.
	bool operator==(const Set& s) const;

  // Used by LeafNode::sort().
  friend int compare(const Set & s1, const Set & s2);
  friend ostream& operator<<(ostream& os, const Set& s);


  // Returns true iff *this is a subset of s.
  bool operator<=(const Set& s) const;
  
  //(Venky): Returns the element e which differentiates "this" from the second set s.
  //That is, this union {e} = s;
  //If not, return -1;
  int ImmediateSubset(const Set& s)
    {
      if (numItems != s.numItems-1)
	return -1;
      
      bool assigned = FALSE;
      int element = -1;
      int index1=0, index2=0;
      for (index1 = 0; index1 < numItems; index1++,index2++)
	{
	  if (items[index1] != s.items[index2])
	    if (FALSE == assigned)
	      {
		element = s.items[index2];
		assigned = TRUE;
		index1--;
	      }
	    else
	      return -1;
	  /***
	  while ((index2 < s.numItems) && (items[index1] != s.items[index2]))
	    {
	      if (index2 == index1)
		{
		  element = s.items[index2];
		  assigned = TRUE;
		}
	      index2++;
	    }
	  if (index2 == s.numItems)
	    return -1;
	  ***/
	}
      if (FALSE == assigned)
	element = s.items[s.numItems-1];
      return element;
    }
  



  private:
	int numItems;
	int activeLow;  // Used by PrefixTree to ignore the first few elements
	int maxNumItems;

	// Array of items. Size is equal to maxNumItems but there
	// are only numItems valid items. maxNumItems increased by
	// NUM_ITEM_STEP when numItems exceeds maxNumItems
	Item * items;
};

#endif
